# coding: utf-8
# 2019/9/20 @ tongshiwei

from longling.Architecture.install_proj.docs_proj import docs_proj
from longling.Architecture.install_file import *


def py_proj(tar_dir, main_params, docs_params):
    variables = {}
    variables.update(main_params)

    pysetup(tar_dir, **variables)

    pytest(tar_dir)

    gitignore("python", tar_dir)

    if docs_params:
        docs_proj(tar_dir, docs_params, __gitignore=False)



# coding: utf-8
# create by tongshiwei on 2019-9-1
__all__ = ["IterableMonitor", "MonitorPlayer", "ProgressMonitor"]

"""
进度监视器，帮助用户知晓当前运行进度 

一个简单的示例如下 ::

    class DemoMonitor(ProgressMonitor):
        def __call__(self, iterator):
            return IterableMonitor(
                iterator,
                self.player, self.player.set_length
            )


    progress_monitor = DemoMonitor(MonitorPlayer())

    for _ in range(5):
        for _ in progress_monitor(range(10000)):
            pass
        print()
"""

from collections import Iterable

try:
    from .stream import flush_print
except (ImportError, SystemError):
    from stream import flush_print

import functools

def pass_function(*args, **kwargs):
    pass


class IterableMonitor(Iterable):
    """
    迭代监控器

    Parameters
    ----------
    iterator:
        待迭代数据
    call_in_iter:
        每次迭代中的回调函数（例如：打印进度等）
    call_after_iter:
        每轮迭代后的回调函数（所有数据遍历一遍后）
    length:
        数据总长（有多少条数据）
    """

    def __init__(self,
                 iterator: (Iterable, list, tuple, dict),
                 call_in_iter=pass_function, call_after_iter=pass_function,
                 length: (int, None) = None,
                 ):
        self.iterator = iter(iterator)
        self.call_in_iter = call_in_iter
        self.call_after_iter = call_after_iter

        self._count = 0
        try:
            if length is not None:
                self._length = length
            else:
                self._length = len(iterator)
        except TypeError:
            self._length = None

    def set_length(self, length):
        self._length = length

    def reset(self):
        self._count = 0
        self._length = None

    def __len__(self):
        return self._length

    def __next__(self):
        try:
            res = self.iterator.__next__()
            self._count += 1
            self.call_in_iter(self._count)
            return res
        except StopIteration:
            self._length = self._count
            self._count = 0
            self.call_after_iter(self._length)
            raise StopIteration

    def __iter__(self):
        return self


class MonitorPlayer(object):
    """监控器显示器"""
    def __init__(self):
        self._count = 0
        self._length = None

    def reset(self):
        self._count = 0
        self._length = None

    def __call__(self, _count):
        self._count = _count
        flush_print("%s|%s" % (self._count, self._length))

    def set_length(self, _length):
        self._length = _length


class ProgressMonitor(object):
    def __init__(self, player):
        self.player = player

    def __call__(self, iterator):
        raise NotImplementedError


if __name__ == '__main__':
    class DemoMonitor(ProgressMonitor):
        def __call__(self, iterator):
            return IterableMonitor(
                iterator,
                self.player, self.player.set_length
            )


    progress_monitor = DemoMonitor(MonitorPlayer())

    for _ in range(5):
        for _ in progress_monitor(range(10000)):
            pass
        print()



from __future__ import absolute_import

from .process_pattern import PatternHitter
from .condition_filter import *
from .variable import *

__all__ = ["PatternHitter", "variable_replace", "default_variable_replace"]



# coding: utf-8
# create by tongshiwei on 2019-8-27

__all__ = [
    "file_ratio_split",
    "train_valid_test", "train_test", "train_valid",
    "kfold"
]
from longling.lib.parser import Formatter, path_append
from longling.lib.stream import wf_open, AddPrinter

from .iterable_splitter import RatioSplitter, KFoldSplitter


def file_ratio_split(filename_list, ratio_list, root_dir=None, filename_formatter_list=None):
    assert len(ratio_list) == len(filename_formatter_list)
    if root_dir is not None:
        filename_formatter_list = [path_append(root_dir, f) for f in filename_formatter_list]
    wfps = []
    fps = []
    source_target_iterable = []
    for filename in filename_list:
        _wfps = [
            wf_open(Formatter.format(filename, formatter=formatter) for formatter in filename_formatter_list)
        ]
        wfps.append(_wfps)
        targets = [
            AddPrinter(wfp) for wfp in _wfps
        ]
        fp = open(filename, encoding="utf-8")
        fps.append(fp)
        source_target_iterable.append((fp, targets))

    splitter = RatioSplitter(fps[0], *ratio_list)
    splitter.split(*source_target_iterable)

    for _wfps in wfps:
        for wfp in _wfps:
            wfp.close()
    for fp in fps:
        fp.close()


def train_valid_test(*filename,
                     train_ratio=0.6, valid_ratio=0.2, test_ratio=0.2,
                     root_dir=None,
                     train_filename="{}.train", valid_filename="{}.valid", test_filename="{}.test"):
    file_ratio_split(
        filename,
        [train_ratio, valid_ratio, test_ratio],
        root_dir,
        [train_filename, valid_filename, test_filename]
    )


def train_valid(*filename: str,
                train_ratio: float = 0.8, valid_ratio: float = 0.2,
                root_dir: str = None,
                train_filename="{}.train", valid_filename="{}.valid"):
    """
    1. filename 可以是一个或多个字符串，指明待切分的文件名，可以通过指定root_dir来批量增加路径前缀

    2. train_filename 和 valid_filename 都是通过字符串format方式生成的:
    $train_filename = train_filename.format($filename),
    其中 $train_filename 指的是生成并使用存储训练数据的文件文件名，
    而 $filename 指的是第一步中合并root_dir之后的源文件文件名

    Parameters
    ----------
    filename: str or list
        需要进行划分的源文件
    train_ratio
    valid_ratio
    root_dir: str
        根目录
    train_filename:

    valid_filename

    Returns
    -------

    """
    file_ratio_split(
        filename,
        [train_ratio, valid_ratio],
        root_dir,
        [train_filename, valid_filename]
    )


def train_test(*filename,
               train_ratio=0.8, test_ratio=0.2,
               root_dir=None,
               train_filename="{}.train", test_filename="{}.valid"):
    file_ratio_split(
        filename,
        [train_ratio, test_ratio],
        root_dir,
        [train_filename, test_filename]
    )


def kfold(*filename, n_splits, root_dir=None, train_filename="{}_{}.train", test_filename="{}_{}.test"):
    if root_dir is not None:
        train_filename = path_append(root_dir, train_filename)
        test_filename = path_append(root_dir, test_filename)

    target_filename_list = [
        [
            [Formatter.format(_filename, i, train_filename) for i in range(n_splits)],
            [Formatter.format(_filename, i, test_filename) for i in range(n_splits)]
        ]
        for _filename in filename
    ]

    wfps = []
    fps = []
    source_target_iterable = []
    for _source_target_iterable in zip(filename, target_filename_list):
        _wfps = [
            wf_open(_wf) for _wf in _source_target_iterable[1]
        ]
        targets = [
            AddPrinter(fp) for fp in _wfps
        ]
        wfps.append(_wfps)
        fp = open(_source_target_iterable[0], encoding="utf-8")
        fps.append(fp)
        source_target_iterable.append((fp, targets))

    splitter = KFoldSplitter(fps[0], n_splits)
    splitter.split(*source_target_iterable)

    for _wfps in wfps:
        for wfp in _wfps:
            wfp.close()
    for fp in fps:
        fp.close()



# coding: utf-8
# 2019/9/20 @ tongshiwei

__all__ = ["cli"]

from longling.lib.stream import build_dir
from longling.Architecture.cli.utils import legal_input
from longling.Architecture.install_proj import project_types
from longling.Architecture.cli.units import *


def cli(skip_top=True, project=None, **kwargs):
    main_params = main_cli(project, **kwargs)

    kwargs.update(main_params)
    if main_params["project_type"] == "docs" or legal_input("Install docs (y/n, default is y) < ",
                                                            __legal_input={'y', 'n'}, __default_value='y') == 'y':
        docs_params = docs_cli(**kwargs)
    else:
        docs_params = {}

    if skip_top:
        tar_dir = "./"
    else:
        tar_dir = build_dir("%s" % main_params["project"])

    __project_type = main_params["project_type"]

    project_types[__project_type](tar_dir=tar_dir, main_params=main_params, docs_params=docs_params)


if __name__ == '__main__':
    cli()



# coding:utf-8
# created by tongshiwei on 2018/10/9

import math
import warnings

import numpy
from longling.ML.toolkit.dataset.splitter.splitter import Splitter
from tqdm import tqdm

__all__ = [
    "RatioSplitter", "TTSplitter", "TVSplitter", "TVTSplitter",
    "KFoldSplitter"

]


class RatioSplitter(Splitter):
    """Splitter for multi ratios, using uniform sampling"""

    def __init__(self, key_iterable, *ratios, shuffle_indices=False, silent=False):
        super(RatioSplitter, self).__init__(key_iterable, shuffle_indices=shuffle_indices, silent=silent)

        ratios = list(ratios)

        ratios[-1] = 1.0 - sum(ratios[:-1]) if ratios[-1] is None else ratios[-1]

        assert isinstance(all(ratios), float), "all of ratios should be float, now are %s" % ratios
        assert 0.0 <= all(ratios) <= 1.0, "all of ratios should be in [0.0, 1.0], now are %s" % ratios

        ratios_sum = sum(ratios)
        if ratios_sum <= 1:
            warnings.warn("the sum of ratios %s is %s, little than 1.0" % (ratios, ratios_sum))
        elif ratios_sum < 0.0 or ratios_sum > 1.0:
            raise ValueError("the sum of ratios %s is %s, not in [0.0, 1.0]" % (ratios, ratios_sum))

        cum_ratios = numpy.cumsum([0.0] + ratios)
        separator_indices = [
            math.ceil(len(self.indices) * ratio) for ratio in cum_ratios
        ]
        self.separator_indices = [
            set(self.indices[separator_indices[i]: separator_indices[i + 1]])
            for i in range(len(separator_indices) - 1)
        ]

    def _split(self, source, target):
        assert len(target) == len(self.separator_indices)
        for idx, elem in tqdm(enumerate(source), "split %s" % source, disable=self.silent):
            for i, separated_indices in enumerate(self.separator_indices):
                if idx in separated_indices:
                    target[i].add(elem)


class TTSplitter(RatioSplitter):
    """Splitter for train and test, using uniform sampling"""

    def __init__(self, key_iterable, train_ratio=0.8, test_ratio=0.2, shuffle_indices=False, **kwargs):
        super(TTSplitter, self).__init__(
            key_iterable, train_ratio, test_ratio, shuffle_indices=shuffle_indices, **kwargs
        )


class TVSplitter(RatioSplitter):
    """Splitter for train and valid, using uniform sampling"""

    def __init__(self, key_iterable, train_ratio=0.8, valid_ratio=0.2, shuffle_indices=False, **kwargs):
        super(TVSplitter, self).__init__(
            key_iterable, train_ratio, valid_ratio, shuffle_indices=shuffle_indices, **kwargs
        )


class TVTSplitter(RatioSplitter):
    """Splitter for train valid and test, using uniform sampling"""

    def __init__(self, key_iterable, train_ratio=0.6, valid_ratio=0.2, test_ratio=0.2, shuffle_indices=False, **kwargs):
        super(TVTSplitter, self).__init__(
            key_iterable, train_ratio, valid_ratio, test_ratio, shuffle_indices=shuffle_indices, **kwargs
        )


class KFoldSplitter(Splitter):
    def __init__(self, key_iterable, n_splits, silent=False):
        super(KFoldSplitter, self).__init__(key_iterable, shuffle_indices=False, silent=silent)
        sample_num = len(self.indices)
        proportion = sample_num / n_splits

        step = math.floor(proportion * sample_num)
        self.indices_buckets = [
            (i, i + step) for i in range(0, sample_num, step)
        ]

    def _split(self, source, target):
        assert len(target) == len(self.indices_buckets)
        for idx, elem in tqdm(enumerate(source), "split %s" % source, disable=self.silent):
            for i, (start, end) in enumerate(self.indices_buckets):
                if start <= idx < end:
                    target[i][0].add(elem)
                else:
                    target[i][1].add(elem)



# coding: utf-8
# 2019/9/21 @ tongshiwei

__all__ = ["main_cli", "docs_cli"]

import datetime
from longling.Architecture.install_proj import project_types
from longling.Architecture.cli.utils import legal_input


def main_cli(project, **kwargs):
    params = dict(
        project=legal_input("Project Name < ", is_legal=lambda x: True if x else False) if not project else project,
        project_type=legal_input("Project Type (%s) < " % "/".join(project_types), __legal_input=project_types),
    )
    return params


def docs_cli(project=None, title=None, author=None, copyright=None, **kwargs):
    params = dict(
        project=legal_input("Project Name < ", is_legal=lambda x: True if x else False) if not project else project,
        docs_root="docs/" if legal_input("mkdir docs? (y/n, default is n) < ",
                                         __legal_input={"y", "n"}, __default_value='n') == "y" else "./"
    )
    params.update(dict(
        title=legal_input("Docs Title (default is %s) < " % params["project"],
                          __default_value=params["project"]) if not title else title,
        author=legal_input("Author < ", is_legal=lambda x: True if x else False) if not author else author,

    ))

    default_copyright = "%s, %s" % (datetime.date.today().year, params["author"])
    params.update(dict(
        copyright=legal_input("Copyright (default is %s) < " % default_copyright,
                              __default_value=default_copyright) if not copyright else copyright,
        readthedocs=True if legal_input("Install .readthedocs.yml (y/n, default is y) < ", __legal_input={'y', 'n'},
                                        __default_value='y') == 'y' else False
    ))

    return params



# coding: utf-8
# 2019/10/12 @ tongshiwei

import json
from longling.lib.candylib import as_list


def key_parser(key):
    if "_" in key:
        # prf_0_f1
        return key.split("_")
    return key


def _select_max(src, *keys):
    keys = as_list(keys)

    result = {
        key: None for key in keys
    }

    with open(src) as f:
        for line in f:
            data = json.loads(line)
            for key in result:
                _key = key_parser(key)
                _data = data
                if isinstance(_key, list):
                    for k in _key:
                        _data = _data[k]
                else:
                    _data = _data[_key]
                if result[key] is None or _data > result[key]:
                    result[key] = _data
    for item in result.items():
        print(item)


def arg_select_max(*keys, src):
    _select_max(src, *keys)


def select_max(src, *keys):
    _select_max(src, *keys)


if __name__ == '__main__':
    arg_select_max("auc", "prf_1_f1", src="result.json")



# coding: utf-8
# create by tongshiwei on 2019-8-30

__all__ =["pick", "tensor2list"]

import torch
from torch import Tensor


def pick(tensor, index, axis=-1):
    return torch.gather(tensor, axis, index.unsqueeze(axis)).squeeze(axis)


def tensor2list(tensor: Tensor):
    return tensor.cpu().tolist()



# coding:utf-8
# created by tongshiwei on 2018/10/9


from .indicesor import Indicesor
from .splitter.file_splitter import *
from .splitter.iterable_splitter import *
from .splitter.splitter import Splitter



# coding: utf-8
# 2019/9/20 @ tongshiwei

from typing import overload

__all__ = ["variable_replace", "default_variable_replace"]

import re

VARIABLE_REGEX = r"(\$[A-z]*)\b"
VARIABLE = re.compile(VARIABLE_REGEX)


@overload
def default_variable_replace(string: str, default_value: dict, key_lower: bool = True, quotation: str = "",
                             **variables) -> str:
    def replace(match):
        variable = string[match.start() + 1:match.end()]
        variable = variable.lower() if key_lower else variable
        if variable not in variables:
            return default_value.get(variable, "")
        return quotation + str(variables[variable]) + quotation

    return VARIABLE.sub(replace, string)


def default_variable_replace(string: str, default_value: str = "", key_lower: bool = True, quotation: str = "",
                             **variables) -> str:
    def replace(match):
        variable = string[match.start() + 1:match.end()]
        variable = variable.lower() if key_lower else variable
        if variable not in variables:
            return default_value
        return quotation + str(variables[variable]) + quotation

    return VARIABLE.sub(replace, string)


def variable_replace(string: str, key_lower: bool = True, quotation: str = "", **variables):
    def replace(match):
        variable = str(variables[string[match.start() + 1:match.end()]])
        variable = variable.lower() if key_lower else variable
        return quotation + variable + quotation

    return VARIABLE.sub(replace, string)



# coding: utf-8
# create by tongshiwei on 2019/7/2

__all__ = ["path_append", "file_exist", "abs_current_dir"]

import os
from pathlib import PurePath


def path_append(path, *addition, to_str=False):
    """
    路径合并函数

    Examples
    --------
    >>> path_append("../", "../data", "../dataset1/", "train")
    PureWindowsPath('../../data/../dataset1/train')

    Parameters
    ----------
    path: str or PurePath
    addition: list(str or PurePath)
    to_str: bool
        Convert the new path to str
    Returns
    -------

    """
    path = PurePath(path) if not isinstance(path, PurePath) else path
    if addition:
        for a in addition:
            path = path / a
    if to_str:
        return str(path)
    return path


def file_exist(file_path):
    """判断文件是否存在"""
    return os.path.isfile(file_path)


def abs_current_dir(file_path):
    """获取文件所在目录的绝对路径"""
    return os.path.abspath(os.path.dirname(file_path))


if __name__ == '__main__':
    import doctest

    doctest.testmod()



# coding: utf-8
# 2019/9/21 @ tongshiwei

from longling.lib.stream import build_dir
from longling.lib.path import path_append
from longling.Architecture.install_file import sphinx_conf, readthedocs, gitignore


def docs_proj(tar_dir, docs_params, __gitignore=True, **kwargs):
    assert docs_params
    variables = {}
    variables.update(docs_params)

    docs_root = build_dir(path_append(tar_dir, docs_params["docs_root"]))
    sphinx_conf(tar_dir=docs_root, **docs_params)

    readthedocs(tar_dir=tar_dir)

    if __gitignore:
        gitignore("docs", tar_dir)



# coding: utf-8
# created by tongshiwei on 18-2-5
import math
import warnings
from collections import OrderedDict
from contextlib import contextmanager

from longling.lib import ProgressMonitor, IterableMonitor
from longling.lib.stream import flush_print

try:
    NAN = math.nan
except (AttributeError, ImportError):
    NAN = float('nan')

__all__ = ["ConsoleProgressMonitor"]


class ConsoleProgressMonitor(ProgressMonitor):
    def __init__(self, indexes: (dict, OrderedDict), values: (dict, None) = None,
                 batch_num=NAN, end_epoch=NAN, output=True, silent=False, ):
        super(ConsoleProgressMonitor, self).__init__(
            ConsoleProgressMonitorPlayer(
                indexes=indexes,
                values=values,
                batch_num=batch_num,
                end_epoch=end_epoch,
                output=output,
                silent=silent,
            )
        )

    def __call__(self, iterator, epoch, *args, **kwargs):
        self.player.batch_start(epoch)
        return IterableMonitor(iterator, self.player, self.player.batch_end)


class ConsoleProgressMonitorPlayer(object):
    def __init__(self,
                 indexes: (dict, OrderedDict), values: (dict, None) = None,
                 batch_num=NAN, end_epoch=NAN, output=True, silent=False,
                 **kwargs):

        if values is not None:
            assert type(indexes) == type(values)

        if isinstance(indexes, dict):
            indexes = OrderedDict(indexes)

        if values is not None:
            for prefix, names in indexes.items():
                for name in names:
                    _ = values[prefix][name]

        self.indexes = indexes

        self.batch_num = batch_num
        self._batch_num = None
        self.end_epoch = end_epoch

        self.output = output

        info_header = "{:>5}| {:>7}" + " " * 5 \
                      + "{:>10}" + " " * 2 + "{:>10}" + " " * 5

        index_header = ""
        arguments = []

        for prefix, _indexes in self.indexes.items():
            __indexes = ["%s-%s" % (prefix, _index) for _index in _indexes]
            arguments += __indexes
            index_header += (" " * 2).join(
                ["{:>%s}" % min(len(index), 15) for index in __indexes]
            )

        self.output_formatter = info_header + index_header

        self.index_prefix = self.output_formatter.format("Epoch", "Total-E",
                                                         "Batch", "Total-B",
                                                         *arguments)

        self.epoch = NAN
        self.silent = silent

        self.values = values

    def __call__(self, batch_no, **kwargs):
        arguments = []

        for prefix, names in self.indexes.items():
            if prefix not in kwargs:
                ref = self.values[prefix]
            else:
                ref = kwargs[prefix]
            for name in names:
                arguments.append(ref[name])

        for prefix, name_value in kwargs.items():
            if prefix not in self.indexes:
                warnings.warn("detect unknown prefix: %s, all arguments will be ignored" % prefix)

        res_str = self.output_formatter.format(self.epoch, self.end_epoch,
                                               batch_no, self.batch_num,
                                               *arguments)

        if self.output and not self.silent:
            flush_print(res_str)

        self._batch_num = batch_no if self._batch_num is None else max(
            [batch_no, self._batch_num]
        )
        return res_str

    def batch_start(self, epoch):
        res_str = self.index_prefix
        self.epoch = epoch

        if self.output and not self.silent:
            print(res_str)

        return res_str

    def batch_end(self, batch_num=None):
        if batch_num is not None:
            self.batch_num = batch_num
        elif self._batch_num is not None:
            self.batch_num = self._batch_num
        if self.output and not self.silent:
            print("")

        return ""

    @contextmanager
    def watching(self, epoch):
        self.batch_start(epoch)
        yield
        self.batch_end()



# coding: utf-8
# 2019/9/21 @ tongshiwei


def legal_input(__promt: str,
                __legal_input: set = None, __illegal_input: set = None, is_legal=None,
                __default_value=None):
    inp = input(__promt)
    __legal_input = __legal_input if __legal_input else set()
    __illegal_input = __illegal_input if __illegal_input else set()

    def __is_legal(__inp):
        if (not __legal_input or __inp in __legal_input) and __inp not in __illegal_input:
            return True
        elif not inp and __default_value:
            return True
        return False

    is_legal = is_legal if is_legal else __is_legal

    while not is_legal(inp):
        inp = input(__promt)
    if not inp and __default_value:
        return __default_value
    return inp



# coding: utf-8
# 2019/9/20 @ tongshiwei

__all__ = ["gitignore", "pytest", "pysetup", "sphinx_conf", "readthedocs"]

import functools
from shutil import copyfile

from longling import wf_open
from longling.lib.path import abs_current_dir, path_append
from longling.lib.process_pattern import default_variable_replace
from longling.lib.utilog import config_logging, LogLevel

logger = config_logging(logger="arch", console_log_level=LogLevel.INFO)

META = path_append(abs_current_dir(__file__), "meta_docs")
default_variable_replace = functools.partial(default_variable_replace, quotation="\'")


def _template_copy(src, tar, **variables):
    with open(src) as f, wf_open(tar) as wf:
        for line in f:
            print(default_variable_replace(line, default_value="", **variables), end='', file=wf)


def gitignore(atype="", tar_dir="./"):
    src = path_append(META, "gitignore", "%s.gitignore" % atype)
    tar = path_append(tar_dir, ".gitignore")
    logger.info("gitignore: copy %s -> %s" % (src, tar))
    copyfile(src, tar)


def pytest(tar_dir="./"):
    src = path_append(META, "pytest.ini")
    tar = path_append(tar_dir, "pytest.ini")
    logger.info("pytest: copy %s -> %s" % (src, tar))
    copyfile(src, tar)


def coverage():
    pass


def pysetup(tar_dir="./", **variables):
    src = path_append(META, "setup.py.template")
    tar = path_append(tar_dir, "setup.py")
    logger.info("pysetup: template %s -> %s" % (src, tar))
    _template_copy(src, tar, **variables)


def sphinx_conf(tar_dir="./", **variables):
    src = path_append(META, "docs/conf.py.template")
    tar = path_append(tar_dir, "conf.py")
    logger.info("sphinx_conf: template %s -> %s" % (src, tar))
    _template_copy(src, tar, **variables)

    src = path_append(META, "docs/.math.json")
    tar = path_append(tar_dir, ".math.json")
    logger.info("sphinx_conf: copy %s -> %s" % (src, tar))
    copyfile(src, tar)


def readthedocs(tar_dir="./"):
    src = path_append(META, "docs/.readthedocs.yml")
    tar = path_append(tar_dir, ".readthedocs.yml")
    logger.info("readthedocs: copy %s -> %s" % (src, tar))
    copyfile(src, tar)


def docker():
    pass


def travis():
    pass


def gitlab_ci():
    pass
